﻿using System;
using System.Linq;
using System.Net;

namespace com.swivelsecure.pinsafe {
    public class IPAddressRange : IComparable<IPAddressRange> {
        private readonly byte[] addressBytes;
        private readonly byte[] maskBytes;

        public IPAddressRange(string range) {
            if (String.IsNullOrEmpty(range)) {
                IsValid = false;
                return;
            }
            IPAddress address;
            string[] parts = range.Split('/');
            if (!IPAddress.TryParse(parts[0], out address)) {
                IsValid = false;
                return;
            }
            addressBytes = address.GetAddressBytes();
            if (parts.Length == 1) {
                // Just a single address.
                maskBytes = new byte[addressBytes.Length];
                for (int idx = 0; idx < maskBytes.Length; idx++) {
                    maskBytes[idx] = 0xff;
                }
            } else {
                if (parts[1].IndexOf('.') > 0 || parts[1].IndexOf(':') >= 0) {
                    IPAddress mask;
                    if (!IPAddress.TryParse(parts[1], out mask)) {
                        IsValid = false;
                        return;
                    }
                    maskBytes = mask.GetAddressBytes();
                    if (maskBytes.Length != addressBytes.Length) {
                        IsValid = false;
                        return;
                    }
                } else {
                    int bits;
                    if (!Int32.TryParse(parts[1], out bits)) {
                        IsValid = false;
                        return;
                    }
                    int byteMask = bits >> 3; // Number of whole bytes
                    int bitMask = bits & 0xf; // Number of extra bits
                    maskBytes = new byte[address.GetAddressBytes().Length];
                    for (int idx = 0; idx < byteMask && idx < maskBytes.Length; idx++) {
                        maskBytes[idx] = 0xff;
                    }
                    if (byteMask < maskBytes.Length && bitMask != 0) {
                        maskBytes[byteMask] = 0;
                        for (int idx = 7; bitMask > 0; idx--, bitMask--) {
                            maskBytes[byteMask] |= (byte) (1 << idx);
                        }
                    }
                    for (int idx = byteMask + 1; idx < maskBytes.Length; idx++) {
                        maskBytes[idx] = 0;
                    }
                }
                for (int idx = 0; idx < maskBytes.Length; idx++) {
                    addressBytes[idx] &= maskBytes[idx];
                }
            }
			IsValid = true;
		}

        public bool IsValid { get; private set; }

        public bool Matches(IPAddress addr) {
            if (!IsValid || addr == null) {
                return false;
            }
            byte[] addrBytes = addr.GetAddressBytes();
            if (addrBytes.Length != addressBytes.Length) {
                return false;
            }
            return !addrBytes.Where((t, idx) => (t & maskBytes[idx]) != addressBytes[idx]).Any();
        }

        internal static bool Equals(IPAddressRange oldIP, IPAddressRange newIP) {
            if (oldIP == null || newIP == null || !oldIP.IsValid || !newIP.IsValid) {
                return false;
            }
            if (oldIP.addressBytes.Length != newIP.addressBytes.Length) {
                return false;
            }
            return !oldIP.addressBytes.Where((t, idx) => t != newIP.addressBytes[idx] || oldIP.maskBytes[idx] != newIP.maskBytes[idx]).Any();
        }

    	public int CompareTo(IPAddressRange other) {
    		if (other == null) {
    			return 1;
    		}
			if (!IsValid) {
				return (other.IsValid) ? -1 : 0;
			}
			if (!other.IsValid) {
				return 1;
			}
			if (addressBytes.Length != other.addressBytes.Length) {
				return addressBytes.Length - other.addressBytes.Length;
			}
			for (int idx = 0; idx < addressBytes.Length; idx ++) {
				if (addressBytes[idx] != other.addressBytes[idx]) {
					return addressBytes[idx] - other.addressBytes[idx];
				}
			}
			for (int idx = 0; idx < maskBytes.Length; idx++) {
				if (maskBytes[idx] != other.maskBytes[idx]) {
					return maskBytes[idx] - other.maskBytes[idx];
				}
			}
    		return 0;
    	}

		public override bool Equals(object obj) {
			if (obj == null || !(obj is IPAddressRange)) {
				return false;
			}
			IPAddressRange other = (IPAddressRange) obj;
			if (!IsValid) {
				return (other.IsValid);
			}
			if (!other.IsValid) {
				return false;
			}
			if (addressBytes.Length != other.addressBytes.Length) {
				return false;
			}
			for (int idx = 0; idx < addressBytes.Length; idx++) {
				if (addressBytes[idx] != other.addressBytes[idx]) {
					return false;
				}
				if (maskBytes[idx] != other.maskBytes[idx]) {
					return false;
				}
			}
			return true;
		}

		public override int GetHashCode() {
			if (!IsValid) {
				return 0;
			}
			int addrHash = new IPAddress(addressBytes).GetHashCode();
			int maskHash = new IPAddress(maskBytes).GetHashCode();
			return addrHash ^ (maskHash << 2);
		}

    	public override String ToString() {
            if (!IsValid) {
                return "";
            }
            IPAddress addr = new IPAddress(addressBytes);
            IPAddress mask = new IPAddress(maskBytes);
            return addr + "/" + mask;
        }
    }
}
