using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Web;
using System.Xml;

namespace com.swivelsecure.pinsafe {

    public class PINsafeRequest {

        // Last request/response
        private XmlDocument responseDoc;

        /// <summary>Gets the last error.</summary>
        public string Error { get; set; }

        /// <summary>Gets the last warnings.</summary>
        public string[] Warnings { get; set; }

        /// <summary>Gets the last XML request sent to the PINSafe server.</summary>
        public String Request { get; private set; }

        /// <value>Gets the last XML response received from the PINsafe server.</value>
        public string Response { get; set; }

        public void CreateRequest(string action) {
            CreateRequest(new Dictionary<string, string> {{"action", action}});
        }

        public void CreateRequest(string action, string secret) {
            CreateRequest(new Dictionary<string, string> { { "action", action }, {"secret", secret} });
        }

        public void CreateRequest(string action, string username, string secret) {
            CreateRequest(new Dictionary<string, string> { { "action", action }, {"username", username}, { "secret", secret } });
        }

        public void CreateRequest(string action, string secret, Dictionary<string, string> values) {
            if (values == null) {
                CreateRequest(action, secret);
                return;
            }
            values["secret"] = secret;
            values["action"] = action;
            CreateRequest(values);
        }

        public void CreateRequest(string action, string username, string secret, Dictionary<string, string> values) {
            if (values == null) {
                CreateRequest(action, username, secret);
                return;
            }
            values["secret"] = secret;
            values["username"] = username;
            values["action"] = action;
            CreateRequest(values);
        }

        public void CreateRequest(Dictionary<string, string> values) {

            StringBuilder buffer = new StringBuilder();
            XmlWriterSettings writerSettings = new XmlWriterSettings { Encoding = Encoding.UTF8 };
            XmlWriter writer = XmlWriter.Create(buffer, writerSettings);

            writer.WriteStartDocument();
            writer.WriteStartElement("SASRequest");
            foreach (KeyValuePair<string, string> kvp in values) {
                writer.WriteElementString(kvp.Key, kvp.Value);
            }
            writer.WriteEndElement();
            writer.Close();
            Request = buffer.ToString();
        }

        public bool ParseResponse() {
            bool result = false;
            // Parse response
            responseDoc = new XmlDocument();
            responseDoc.LoadXml(Response);
            XmlNodeList resultNodes = responseDoc.GetElementsByTagName("Result");

            if ((resultNodes.Count == 1) && resultNodes[0].InnerText.Equals("PASS")) {
                result = true;

                // Parse out any warnings
                XmlNodeList warningNodes = responseDoc.GetElementsByTagName("Warning");
                Warnings = new string[warningNodes.Count];
                for (int i = 0; i < warningNodes.Count; i++) {
                    Warnings[i] = warningNodes[i].InnerText;
                }
            } else {
                // Parse out the error
                XmlNodeList errorNodes = responseDoc.GetElementsByTagName("Error");
                if (errorNodes.Count > 0) {
                    Error = errorNodes[0].InnerText;
                }
            }
            return result;
        }

        public string GetResponseField(string name) {
            XmlNodeList nodes = responseDoc.GetElementsByTagName(name);
            return (nodes.Count > 0) ? nodes[0].InnerText : "";
        }
    }

    /// <summary><c>PINsafeClient</c> provides convenient access to the
    ///		authentication service of a PINsafe server</summary>
    public class PINsafeClient {

		public const string PIN_EXPIRES_SOON = "AGENT_WARN_PIN_EXPIRY";
		public const string PIN_EXPIRES_TODAY = "AGENT_WARN_PIN_EXPIRES_TODAY";
		public const string PIN_EXPIRED = "AGENT_ERROR_PIN_EXPIRED";

        // Server settings

        private readonly PINsafeSettings settings;

        private static readonly PINsafeClient webInstance = new PINsafeClient();

        public static PINsafeClient GetWebInstance() {
            return webInstance;
        }

        private PINsafeClient() {
            settings = new PINsafeSettings();
        }

        public PINsafeClient(PINsafeSettings settings) {
            this.settings = settings;
        }

        public PINsafeSettings Settings {
            get { return settings; }
        }

        /// <summary>Gets or sets the secret used for communicating with the
        ///		PINsafe server.</summary>
        private string Secret {
            get { return settings.Secret; }
        }

        private bool AcceptSelfSigned {
            get { return settings.AcceptSelfSigned; }
        }

        private static string CookieSecret {
            get { return PINsafeSettings.CookieSecret; }
        }

        private bool AllowNonPINsafeUsers {
            get { return settings.AllowNonPINsafeUsers; }
        }

        /// <summary>The URL for accessing the PINsafe server.</summary>
        public string Address {
            get {
                return GetServletAddress("AgentXML");
            }
        }

        /// <summary>The URL for accessing the PINsafe server.</summary>
/*        private string ImageAddress {
            get { return GetServletAddress("SCImage"); }
        }*/

        /// <summary>The URL for accessing the PINsafe server.</summary>
        private string GetServletAddress(String servlet) {
            return GetServletAddress(settings.ServerUrl, servlet);
        }

		/// <summary>The URL for accessing the PINsafe server.</summary>
		private string GetProxyServletAddress(String servlet) {
			return GetServletAddress(settings.ProxyServerUrl, servlet);
		}

        private static string GetServletAddress(String serverUrl, String servlet) {
            StringBuilder addressBuilder = new StringBuilder();

            addressBuilder.Append(serverUrl);

            addressBuilder.Append(servlet);

            return addressBuilder.ToString();
        }

        /// <summary>The URL for accessing the PINsafe server.</summary>
/*        private string MessageAddress {
            get { return GetServletAddress("DCMessage"); }
        }*/

        /// <summary>
        /// Check whether a user exists or not.
        /// </summary>
        /// <param name="request"></param>
        /// <param name="username"></param>
        /// <returns></returns>
        public bool UserExists(PINsafeRequest request, string username) {
            try {
                // Create XML
                string adjustUsername = AdjustUsername(username);
                request.CreateRequest("exists", adjustUsername, Secret);

                bool result = DoRequest(request);
                Debug.WriteLine(String.Format("UserExists: user {0} {1}", adjustUsername, result ? "exists" : "does not exist"));
                return result;
            } catch (Exception e) {
                request.Error = e.Message;
                return false;
            }
        }

        public string UserExistsByAttribute(PINsafeRequest request, string username) {
            try {
                // Create XML
                request.CreateRequest("existsbyattribute", AdjustUsername(username), Secret);

                if (DoRequest(request)) {
                    string attr = request.GetResponseField("AttributeName");
                    Debug.WriteLine(String.IsNullOrEmpty(attr)
                                        ? String.Format("UserExistsByAttribute: User {0} not found", username)
                                        : String.Format("UserExistsByAttribute: User {0} found with attribute {1}",
                                                        username, attr));
                    return attr;
                }
                Debug.WriteLine(String.Format("UserExistsByAttribute: Request failed for user {0}", username));
                return "";
            } catch (Exception e) {
                Debug.WriteLine(String.Format("UserExistsByAttribute: Error checking user {0}: {1}", username, e.Message));
                request.Error = e.Message;
                return "";
            }
        }

        public string GetUsernameByAttribute(PINsafeRequest request, string attribute, string userattr) {
            try {
                // Create XML
                request.CreateRequest("getusernamebyattribute", Secret, new Dictionary<string, string> {{"AttributeValue", AdjustUsername(userattr)},{"AttributeName", attribute}});

                if (DoRequest(request)) {
                    return request.GetResponseField("Username");
                }

                return "";
            } catch (Exception e) {
                Debug.WriteLine("Error getting username: " + e.Message);
                request.Error = e.Message;
                return "";
            }
        }

        /// <summary>Starts a new single channel session with the PINsafe 
        ///		server.</summary>
        ///<param name="request"></param>
        ///<param name="userAttr">The username for the user.</param>
        ///<returns>Session ID for the started session, or an empty string if 
        ///		an error occurred.</returns>
        public string StartSession(PINsafeRequest request, string userAttr) {
            string sessionId = StartSessionByUsername(request, userAttr);
            if (settings.EnableMultipleUsernames && string.IsNullOrEmpty(sessionId)) {
                string attribute = UserExistsByAttribute(request, userAttr);
                if (!string.IsNullOrEmpty(attribute)) {
                    string username = GetUsernameByAttribute(request, attribute, userAttr);
                    if (!string.IsNullOrEmpty(username)) {
                        sessionId = StartSessionByUsername(request, username);
                    }
                }
            }
            return sessionId;
        }

        private string StartSessionByUsername(PINsafeRequest request, string username) {
            string sessionID;

            try {
                // Create XML
                request.CreateRequest("sessionstart", AdjustUsername(username), Secret);

                sessionID = DoRequest(request) ? request.GetResponseField("SessionID") : "";
            } catch (Exception e) {
                request.Error = e.Message;
                sessionID = "";
            }

            return sessionID;
        }

        /// <summary>Perform an authentication with the PINsafe server.</summary>
        ///<param name="request"></param>
        ///<param name="username">The username of the user.</param>
        ///<param name="password">The user's password.</param>
        ///<param name="OTC">The user's one time code.</param>
        ///<param name="expiryTime">The time at which the login cookie will expire.</param>
        ///<returns>True if the authentication was successful, otherwise 
        ///		false.</returns>
        public string Login(PINsafeRequest request, string username, string password, string OTC, DateTime expiryTime) {
            bool result;

            try {
                // If Non-PINsafe users are allowed, check if the user exists.
                bool doLogin = true;
                if (AllowNonPINsafeUsers) {
                    doLogin = UserExists(request, username);
                }

                // Create XML
                if (doLogin) {
                    string adjustUsername = AdjustUsername(username);
                    request.CreateRequest("login", adjustUsername, Secret, new Dictionary<string, string>{{"password", password},{"otc", OTC}});
                    // Attempt authentication
                    result = DoRequest(request);
                    Debug.WriteLine(String.Format("Login: PINsafe login for user {0} {1}", adjustUsername, result ? "succeeded" : "failed"));
                } else {
                    // By-pass PINsafe login.
                    Debug.WriteLine(String.Format("Login: PINsafe login for user {0} not required", username));
                    result = true;
                }
            } catch (Exception e) {
                request.Error = e.Message;
                Debug.WriteLine(String.Format("Login: Error in PINsafe login for user {0}: {1}", username, e.Message));
                result = false;
            }

            return (result) ? GetCookie(username, expiryTime) : null;
        }

        /// <summary>Perform an authentication with the PINsafe server.</summary>
        /// <param name="request"></param>
        /// <param name="username">The username of the user.</param>
        /// <param name="attribute">The attribute name for the username parameter.</param>
        /// <param name="password">The user's password.</param>
        /// <param name="otc">The user's one time code.</param>
        /// <param name="expiryTime">The time at which the login cookie will expire.</param>
        /// <returns>True if the authentication was successful, otherwise 
        ///		false.</returns>
        public string LoginByAttribute(PINsafeRequest request, string username, string attribute, string password, string otc, DateTime expiryTime) {
            bool result;

            try {

                // Create XML
                string adjustUsername = AdjustUsername(username);
                request.CreateRequest("loginbyattribute", adjustUsername, Secret, new Dictionary<string, string> {{"attribute", attribute}, {"password", password}, {"otc", otc}});

                // Attempt authentication
                result = DoRequest(request);
                Debug.WriteLine(String.Format("LoginByAttribute: PINsafe login for user {0} {1}", adjustUsername, result ? "succeeded" : "failed"));
            } catch (Exception e) {
                request.Error = e.Message;
                Debug.WriteLine(String.Format("LoginByAttribute: Error in PINsafe login for user {0}: {1}", username, e.Message));
                result = false;
            }

            return (result) ? GetCookie(username, expiryTime) : null;
        }

        public string GetCookie(string username, DateTime expiryTime) {
            if (String.IsNullOrEmpty(username)) {
                return null;
            }
            TimeSpan expirySpan = expiryTime.Subtract(new DateTime(2000, 1, 1));
            string expirySecs = ((long) expirySpan.TotalSeconds).ToString("D10");
            string userid = username.ToLower(CultureInfo.InvariantCulture);

            return GetHashValue(expirySecs + userid) + userid + userid.Length.ToString("D3") + expirySecs;
        }

        public string UpdateCookie(string oldCookie) {
            if (CookieValid(oldCookie)) {
                int cookieLen = oldCookie.Length;
                string expiryString = oldCookie.Substring(cookieLen - 10);
                string sNameLength = oldCookie.Substring(cookieLen - 13, 3);
                int nameLength = int.Parse(sNameLength);
                string username = oldCookie.Substring(cookieLen - 13 - nameLength, nameLength);
                DateTime oldExpiryTime = new DateTime(2000, 1, 1).AddSeconds(long.Parse(expiryString));
                TimeSpan timeToExpire = oldExpiryTime - DateTime.Now;
                if (timeToExpire.TotalSeconds < settings.IdleTimeSecs - 10) {
                    // Cookie is more than 10 seconds old - renew it.
                    DateTime expiryTime = DateTime.Now.AddSeconds(settings.IdleTimeSecs);
                    return GetCookie(username, expiryTime);
                }
            }
            return null;
        }

        private string GetHashValue(string timeAndUser) {
            return GetHashValue(CookieSecret, timeAndUser);
        }

        private string GetHashValue(string secret, string timeAndUser) {
            string data = secret + timeAndUser;
            byte[] bytes = SHA512.Create().ComputeHash(Encoding.UTF8.GetBytes(data));
            StringBuilder hash = new StringBuilder();
            foreach (byte b in bytes) {
                hash.Append(b.ToString("X2"));
            }
            return hash.ToString();
        }

        public bool CookieValid(string cookieData) {
            // Split the cookie up.
            int cookieLen = (cookieData == null) ? 0 : cookieData.Length;
            if (cookieLen <= 16) {
                Debug.WriteLine("CookieValid failed: cookie too short");
                return false;
            }
            int hashLen = cookieLen - 10;
// ReSharper disable PossibleNullReferenceException
	        string expiryString = cookieData.Substring(hashLen);
// ReSharper restore PossibleNullReferenceException
            long expiryLong;
            if (!long.TryParse(expiryString, out expiryLong)) {
                Debug.WriteLine("CookieValid failed: invalid expiry string: " + expiryString);
                return false;
            }
            hashLen -= 3;
            string sNameLength = cookieData.Substring(hashLen, 3);
            int nameLength;
            if (!int.TryParse(sNameLength, out nameLength)) {
                Debug.WriteLine("CookieValid failed: invalid name length: " + sNameLength);
                return false;
            }
            hashLen -= nameLength;
            if (hashLen <= 0) {
                Debug.WriteLine("CookieValid failed: cookie too short");
                return false;
            }
            string username = cookieData.Substring(hashLen, nameLength);
            DateTime expiryTime = new DateTime(2000, 1, 1).AddSeconds(expiryLong);
            if (expiryTime < DateTime.Now) {
                Debug.WriteLine("CookieValid failed: cookie has expired");
                return false; // Cookie has expired.
            }
            string orgHash = cookieData.Substring(0, hashLen);
            string timeAndUser = expiryString + username.ToLower(CultureInfo.InvariantCulture);
            if (String.CompareOrdinal(orgHash, GetHashValue(timeAndUser)) == 0) {
                return true;
            }
            if (String.CompareOrdinal(orgHash, GetHashValue(PINsafeSettings.LastCookieSecret, timeAndUser)) == 0) {
                return true;
            }
            Debug.WriteLine("CookieValid failed: hash value mismatch");
            return false;
        }

        /// <summary>Perform a PIN change with the PINsafe server.</summary>
        ///<param name="request"></param>
        ///<param name="username">The username of the user.</param>
        ///<param name="password">The user's current password.</param>
        ///<param name="newPassword">The user's new password.</param>
        ///<param name="OTC">The one time code for the user's current 
        ///  PIN.</param>
        ///<param name="newOTC">The one time code for the user's new 
        ///  PIN.</param>
        ///<returns>True if the PIN change was successful, otherwise 
        ///		false</returns>
        public bool ChangePin(PINsafeRequest request, string username, string password, string newPassword,
                              string OTC, string newOTC) {
            bool result;

            try {
                // Create XML
                request.CreateRequest("changepin", AdjustUsername(username), Secret,
                                      new Dictionary<string, string> {
                                                                         {"password", password},
                                                                         {"newpassword", newPassword},
                                                                         {"otc", OTC},
                                                                         {"newotc", newOTC}
                                                                     });

                // Attempt pin change
                result = DoRequest(request);
            } catch (Exception e) {
                request.Error = e.Message;
                result = false;
            }

            return result;
        }

        /// <summary>Send the request to the PINsafe server and parse the result.</summary>
        /// <returns>True if the request was successful, false otherwise.</returns>
        private bool DoRequest(PINsafeRequest request) {
            try {
                return DoRequest(request, settings.ServerUrl);
            } catch (Exception) {
                return CheckServers(false) && DoRequest(request, settings.ServerUrl);
            }
        }

        /// <summary>Send the request to the PINsafe server and parse the result.</summary>
        /// <returns>True if the request was successful, false otherwise.</returns>
        private bool DoRequest(PINsafeRequest request, string address) {
            SendRequest(request, GetServletAddress(address, "AgentXML"));
            return request.ParseResponse();
        }

        private void SendRequest(PINsafeRequest request, string address) {
            request.Response = "";

            byte[] requestData = Encoding.UTF8.GetBytes(request.Request);

            WebClient client = new WebClient();
            if (AcceptSelfSigned) {
                ServicePointManager.ServerCertificateValidationCallback = DummyCertificateValidator;
            }
            byte[] responseData = client.UploadData(address, requestData);

            request.Response = Encoding.UTF8.GetString(responseData);
        }

        private bool DummyCertificateValidator(object sender, X509Certificate certificate, X509Chain chain,
                                               SslPolicyErrors sslPolicyErrors) {
            return true;
        }

        public static Image BlankImage {
            get {
                Bitmap img = new Bitmap(1, 1);
                img.SetPixel(0, 0, Color.White);
                return img;
            }
        }

        private Image GetImage(PINsafeRequest request, String sessionID, String username, string servletUrl, out int status) {
            if (string.IsNullOrEmpty(sessionID)) {
                if (!string.IsNullOrEmpty(username)) {
                    sessionID = StartSession(request, username);
                } else {
                    status = (int) HttpStatusCode.OK;
                    return BlankImage;
                }
            }

            if (string.IsNullOrEmpty(sessionID)) {
                status = (int) HttpStatusCode.OK;
                return BlankImage;
            }

            StringBuilder address = new StringBuilder(GetProxyServletAddress(servletUrl));
            address.Append("?sessionid=");
            address.Append(sessionID);
            address.Append("&random=");
            address.Append(new Random().Next());
            string contentType;
            HttpStatusCode statusCode;
            Image image = GetPinsafeResponse(address.ToString(), out contentType, out statusCode);
            status = (int) statusCode;
            return statusCode != HttpStatusCode.OK ? BlankImage : image;
        }

        public Image GetImage(PINsafeRequest request, String sessionID, String username, out int status) {
            try {
                return GetImage(request, sessionID, username, "SCImage", out status);
            } catch (Exception) {
                status = (int) HttpStatusCode.NotFound;
                return BlankImage;
            }
        }

        public Image GetMessage(PINsafeRequest request, String sessionID, String username, out int status) {
            try {
                return GetImage(request, sessionID, username, "DCMessage", out status);
            } catch (Exception) {
                status = (int)HttpStatusCode.NotFound;
                return BlankImage;
            }
        }

		/// <summary>
		/// TODO: This should probably return a proper blank digit.
		/// </summary>
		public static Image BlankPinpadImage {
			get {
				Bitmap img = new Bitmap(1, 1);
				img.SetPixel(0, 0, Color.White);
				return img;
			}
		}

		public Image GetPinpadImage(PINsafeRequest request, String username, String padno, out int status) {
			try {
				StringBuilder address = new StringBuilder(GetProxyServletAddress("SCPinPad"));
				address.Append("?username=");
				address.Append(AdjustUsername(username));
				address.Append("&padno=");
				address.Append(padno);
				address.Append("&random=");
				address.Append(new Random().Next());
				string contentType;
				HttpStatusCode statusCode;
				Image image = GetPinsafeResponse(address.ToString(), out contentType, out statusCode);
				status = (int)statusCode;
				return statusCode != HttpStatusCode.OK ? BlankImage : image;
			} catch (Exception) {
				status = (int)HttpStatusCode.NotFound;
				return BlankPinpadImage;
			}
		}

        /// <summary>
        /// Send a request to the PINSafe server, and get the image data.
        /// </summary>
        /// <param name="requestUrl">The URL of the PINSafe request.</param>
        /// <param name="contentType">The content type of the returned image.</param>
        /// <param name="status">The response code from the PINsafe server.</param>
        /// <returns>The PINSafe image data.</returns>
        private Image GetPinsafeResponse(string requestUrl, out string contentType, out HttpStatusCode status) {
            if (AcceptSelfSigned) {
                ServicePointManager.ServerCertificateValidationCallback = DummyCertificateValidator;
            }
            ServicePointManager.Expect100Continue = false;
            HttpWebRequest webRequest = (HttpWebRequest) WebRequest.Create(requestUrl);
            HttpWebResponse webResponse = (HttpWebResponse) webRequest.GetResponse();
            status = webResponse.StatusCode;
            if (status == HttpStatusCode.OK) {
                MemoryStream strm = new MemoryStream();
                Stream respStream = webResponse.GetResponseStream();
                if (respStream != null) {
                    const int bufSize = 0x400;
                    byte[] buffer = new byte[bufSize];
                    int bufLen = respStream.Read(buffer, 0, bufSize);
                    while (bufLen > 0) {
                        strm.Write(buffer, 0, bufLen);
                        bufLen = respStream.Read(buffer, 0, bufSize);
                    }
                }
                contentType = webResponse.ContentType;
                return Image.FromStream(strm);
            }
            contentType = "";
            return null;
        }

        public string InitialiseOutlookSession(string logonPath, string userAgent, out string error) {
            error = "";
            HttpWebRequest owaRequest = (HttpWebRequest) WebRequest.Create(logonPath);
            owaRequest.Method = "GET";
            owaRequest.Accept = "*/*";
            owaRequest.Headers["Accept-Language"] = "en-gb";
            owaRequest.UserAgent = userAgent;
            owaRequest.CookieContainer = new CookieContainer();
            HttpWebResponse owaResponse;
            try {
                owaResponse = (HttpWebResponse) owaRequest.GetResponse();
            } catch (WebException e) {
                owaResponse = (HttpWebResponse) e.Response;
                if (owaResponse == null) {
                    error = e.Message + "<br/>" + e.StackTrace;
                } else {
                    error = "Exception thrown by authentication request: " + owaResponse.StatusCode + " - " + e.Message;
                    try {
                        using (Stream respStrm = owaResponse.GetResponseStream()) {
                            if (respStrm != null) {
                                StreamReader reader = new StreamReader(respStrm);
                                string respText = reader.ReadToEnd();
                                error += "<br/>" + respText;
                            }
                        }
                    } catch (Exception) {
                        error += "<br/>Unable to read response";
                    }
                }
                return null;
            }
            Cookie sessionCookie = owaResponse.Cookies["OutlookSession"];
            return sessionCookie == null ? null : sessionCookie.Value;
        }

        public CookieCollection LoginOWA(string authPath, string username, string password, HttpCookieCollection cookies,
                                         string host, ref string destination, int flags, int trustLevel,
                                         string userAgent, out string error) {
            error = "";
            if (AcceptSelfSigned) {
                ServicePointManager.ServerCertificateValidationCallback = DummyCertificateValidator;
            }
            ServicePointManager.Expect100Continue = false;
            StringBuilder req = new StringBuilder();
            req.Append("destination=");
            req.Append(HttpUtility.UrlEncode(destination));
            req.Append("&flags=");
            req.Append(flags);
            req.Append("&forcedownlevel=0&trusted=");
            req.Append(trustLevel);
            req.Append("&username=");
            req.Append(HttpUtility.UrlEncode(username));
            req.Append("&password=");
            req.Append(HttpUtility.UrlEncode(password));
            req.Append("&isUtf8=1");
            HttpWebRequest owaRequest = (HttpWebRequest) WebRequest.Create(authPath);
            owaRequest.Method = "POST";
            owaRequest.ContentType = "application/x-www-form-urlencoded";
            owaRequest.UserAgent = userAgent;
            CookieContainer cookieJar = new CookieContainer();
            owaRequest.CookieContainer = cookieJar;
            if (cookies != null) {
                foreach (string name in cookies) {
                    HttpCookie cookie = cookies[name];
                    if (cookie == null) {
                        continue;
                    }
                    string value = cookie.Value;
                    owaRequest.CookieContainer.Add(new Uri(authPath), new Cookie(name, value));
                }
            }
            owaRequest.AllowAutoRedirect = false; // Don't follow the redirect.
            using (Stream stream = owaRequest.GetRequestStream()) {
                byte[] reqBytes = Encoding.UTF8.GetBytes(req.ToString());
                //owaRequest.ContentLength = reqBytes.Length;
                stream.Write(reqBytes, 0, reqBytes.Length);
                stream.Close();
            }
            HttpWebResponse owaResponse;
            try {
                owaResponse = (HttpWebResponse) owaRequest.GetResponse();
            } catch (WebException e) {
                owaResponse = (HttpWebResponse) e.Response;
                if (owaResponse == null) {
                    error = e.Message + "<br/>" + e.StackTrace;
                } else {
                    error = "Exception thrown by authentication request: " + owaResponse.StatusCode + " - " + e.Message;
                    try {
                        using (Stream respStrm = owaResponse.GetResponseStream()) {
                            if (respStrm != null) {
                                StreamReader reader = new StreamReader(respStrm);
                                string respText = reader.ReadToEnd();
                                error += "<br/>" + respText;
                            }
                        }
                    } catch (Exception) {
                        error += "<br/>Unable to read response";
                    }
                }
                return null;
            }
            if (owaResponse.StatusCode != HttpStatusCode.OK && owaResponse.StatusCode != HttpStatusCode.Redirect) {
                error = owaResponse.StatusCode + ": " + owaResponse.StatusDescription;
                return null;
            }
            if (owaResponse.StatusCode == HttpStatusCode.Redirect) {
                destination = owaResponse.Headers[HttpResponseHeader.Location];
            }
            CookieCollection respCookies = owaResponse.Cookies;
            if (respCookies == null || respCookies.Count < 2) {
                error = "OWA login failed: insufficient cookies returned";
                return null;
            }
            owaResponse.Close();
            return respCookies;
        }

        private string AdjustUsername(string username) {
            if (username == null) {
                return null;
            }
            if (settings.IgnoreDomainPrefix) {
                int slashPos = username.IndexOf("\\", StringComparison.Ordinal);
                if (slashPos > 0 && slashPos < username.Length) {
                    return username.Substring(slashPos + 1);
                }
            }
            if (settings.IgnoreDomainSuffix) {
                int atPos = username.IndexOf('@');
                if (atPos > 0) {
                    return username.Substring(0, atPos);
                }
            }
            return username;
        }

        public bool CheckAuthenticationRequired(PINsafeRequest request, string username, string clientIP) {
            if (!settings.FilterEnabled) {
                // Filter disabled: no authentication required.
                Debug.WriteLine("Filter disabled: no authentication required");
                return false;
            }
            if (settings.AllowNonPINsafeUsers) {
                if (settings.EnableMultipleUsernames) {
                    if (string.IsNullOrEmpty(UserExistsByAttribute(request, username))) {
                        return false;
                    }
                } else if (!UserExists(request, username)) {
                    return false;
                }
            }
            return CheckClientIP(clientIP);
        }

        public bool CheckClientIP(string clientIP) {
            // check if the client IP requires PINsafe authentication.
            IPAddress clientAddr;
            if (!IPAddress.TryParse(clientIP, out clientAddr)) {
                // Shouldn't happen, but let's be safe.
                return settings.IncludeIPs;
            }
            bool matchesIP = (settings.ExcludedIPs.Any(range => range.Matches(clientAddr)));
            return !(matchesIP ^ settings.IncludeIPs);
        }

        public bool Ping(PINsafeRequest request, string address) {
            request.CreateRequest("ping");
            try {
                return DoRequest(request, address);
            } catch (Exception) {
                return false;
            }
        }

        private bool TestServer(ServerSettings server) {
            try {
                return Ping(new PINsafeRequest(), server.Address);
            } catch (WebException) {
                return false;
            }
        }

        public bool CheckServers(bool checkCurrent=true) {
            if (checkCurrent) {
                if (TestServer(settings.ActiveServer)) {
                    return true;
                }
            }
            for (int idx=0; idx<settings.ServerCount; idx++) {
                ServerSettings server = settings.GetServer(idx);
                if (server == settings.ActiveServer) continue;
                if (!TestServer(server)) continue;
                settings.ActiveServerId = idx;
                return true;
            }
            return false;
        }
    }
}
