using System;
using System.Collections.Generic;
using System.Configuration;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml;
using Microsoft.Win32;

namespace com.swivelsecure.pinsafe {

    /// <summary>
    /// Settings for a single server.
    /// </summary>
    public class ServerSettings {
        private const string PINSAFE_SERVER = "PINsafeServer";
        private const string PINSAFE_SECURE = "PINsafeSecure";
        private const string PINSAFE_PORT = "PINsafePort";
        private const string PINSAFE_CONTEXT = "PINsafeContext";
        private const string PINSAFE_SECRET = "PINsafeSecret";
        private const string PINSAFE_ACCEPT_SELFSIGNED = "PINsafeAcceptSelfSigned";
		private const string PINSAFE_PROXY_SERVER = "PINsafeProxyServer";
		private const string PINSAFE_PROXY_SECURE = "PINsafeProxySecure";
		private const string PINSAFE_PROXY_PORT = "PINsafeProxyPort";
		private const string PINSAFE_PROXY_CONTEXT = "PINsafeProxyContext";

		public ServerSettings() {
			SetDefaults();
		}

		public void SetDefaults() {
			Server = "";
			Port = 8080;
			Context = "pinsafe";
			Ssl = true;
			ProxyServer = "";
			ProxyPort = 8443;
			ProxyContext = "proxy";
			ProxySsl = true;
			AcceptSelfSigned = false;
			Secret = "";
		}

        public bool Ssl { get; set; }

        public string Server { get; set; }

        public int Port { get; set; }

        public string Context { get; set; }

        public string Secret { get; set; }

        public bool AcceptSelfSigned { get; set; }

		public bool ProxySsl { get; set; }

		public string ProxyServer { get; set; }

		public int ProxyPort { get; set; }

		public string ProxyContext { get; set; }

        public string Address {
			get { return GetAddress(Ssl, Server, Port, Context); }
		}

		public string ProxyAddress {
			get { return GetAddress(ProxySsl, ProxyServer, ProxyPort, ProxyContext); }
		}

		private static string GetAddress(bool ssl, string server, int port, string context) {
			StringBuilder addressBuilder = new StringBuilder();

			addressBuilder.Append(ssl ? "https://" : "http://");

			addressBuilder.AppendFormat("{0}:{1}/{2}/", server, port, context);

			return addressBuilder.ToString();			
		}

        public bool LoadSettings(string suffix) {
            Server = ConfigurationManager.AppSettings[PINSAFE_SERVER+suffix];
            if (String.IsNullOrEmpty(Server)) {
                return false;
            }
            String sPort = ConfigurationManager.AppSettings[PINSAFE_PORT + suffix];
            Context = ConfigurationManager.AppSettings[PINSAFE_CONTEXT + suffix];
            Secret = ConfigurationManager.AppSettings[PINSAFE_SECRET + suffix];
            String sSsl = ConfigurationManager.AppSettings[PINSAFE_SECURE + suffix];
            String allowSelfSigned = ConfigurationManager.AppSettings[PINSAFE_ACCEPT_SELFSIGNED + suffix];
            if (sSsl == null || sSsl.Equals("")) {
                Ssl = false;
            } else {
                bool ssl;
                Ssl = Boolean.TryParse(sSsl, out ssl) && ssl;
            }
            if (sPort == null || sPort.Equals("")) {
                Port = 8080;
            } else {
                int port;
                Port = Int32.TryParse(sPort, out port) ? port : 8080;
            }
            bool acceptSelfSigned;
            AcceptSelfSigned = Boolean.TryParse(allowSelfSigned, out acceptSelfSigned) && acceptSelfSigned;
			ProxyServer = ConfigurationManager.AppSettings[PINSAFE_PROXY_SERVER + suffix];
			if (String.IsNullOrEmpty(ProxyServer)) {
				ProxyServer = Server;
			}
			sPort = ConfigurationManager.AppSettings[PINSAFE_PROXY_PORT + suffix];
			if (sPort == null || sPort.Equals("")) {
				ProxyPort = Port;
			} else {
				int port;
				ProxyPort = Int32.TryParse(sPort, out port) ? port : 8443;
			}
			ProxyContext = ConfigurationManager.AppSettings[PINSAFE_PROXY_CONTEXT + suffix];
			if (String.IsNullOrEmpty(ProxyContext)) {
				ProxyContext = Context;
			}
			sSsl = ConfigurationManager.AppSettings[PINSAFE_PROXY_SECURE + suffix];
			if (sSsl == null || sSsl.Equals("")) {
				ProxySsl = Ssl;
			} else {
				bool ssl;
				ProxySsl = Boolean.TryParse(sSsl, out ssl) ? ssl : Ssl;
			}
			return true;
        }

        public bool LoadSettings(Dictionary<string, string> settings, string suffix) {

            Server = PINsafeSettings.GetSetting(settings, PINSAFE_SERVER + suffix);
            if (String.IsNullOrEmpty(Server)) {
                return false;
            }
            Ssl = PINsafeSettings.GetBoolSetting(settings, PINSAFE_SECURE + suffix, false);
            Port = PINsafeSettings.GetIntSetting(settings, PINSAFE_PORT + suffix, 8080);
            Context = PINsafeSettings.GetSetting(settings, PINSAFE_CONTEXT + suffix);
            Secret = PINsafeSettings.GetSetting(settings, PINSAFE_SECRET + suffix);
            AcceptSelfSigned = PINsafeSettings.GetBoolSetting(settings, PINSAFE_ACCEPT_SELFSIGNED + suffix, false);
			ProxyServer = PINsafeSettings.GetSetting(settings, PINSAFE_PROXY_SERVER + suffix);
			if (String.IsNullOrEmpty(ProxyServer)) {
				ProxyServer = Server;
			}
			ProxySsl = PINsafeSettings.GetBoolSetting(settings, PINSAFE_PROXY_SECURE + suffix, Ssl);
			ProxyPort = PINsafeSettings.GetIntSetting(settings, PINSAFE_PROXY_PORT + suffix, Port);
			ProxyContext = PINsafeSettings.GetSetting(settings, PINSAFE_PROXY_CONTEXT + suffix);
			if (String.IsNullOrEmpty(ProxyContext)) {
				ProxyContext = Context;
			}

            return true;
        }

        public void SaveSettings(Dictionary<string, string> settings, string suffix) {
            settings[PINSAFE_SERVER + suffix] = Server;
            settings[PINSAFE_PORT + suffix] = Port.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_CONTEXT + suffix] = Context;
            settings[PINSAFE_SECRET + suffix] = Secret;
            settings[PINSAFE_SECURE + suffix] = Ssl.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_ACCEPT_SELFSIGNED + suffix] = AcceptSelfSigned.ToString(CultureInfo.InvariantCulture);
			settings[PINSAFE_PROXY_SERVER + suffix] = ProxyServer;
			settings[PINSAFE_PROXY_PORT + suffix] = ProxyPort.ToString(CultureInfo.InvariantCulture);
			settings[PINSAFE_PROXY_CONTEXT + suffix] = ProxyContext;
			settings[PINSAFE_PROXY_SECURE + suffix] = ProxySsl.ToString(CultureInfo.InvariantCulture);
        }
    }

	/// <summary>
	/// Summary description for PINsafeSettings
	/// </summary>
	public class PINsafeSettings {
        public const string DEFAULT_OWA_URL = "/owa/";
	    public const string DEFAULT_AUTH_URL = DEFAULT_OWA_URL + "auth.owa";
	    public const string DEFAULT_AUTH_FOLDER = DEFAULT_OWA_URL + "auth/";
        private const string PINSAFE_MODULE_TYPE = "com.swivelsecure.owafilter.PINsafeOWAFilter";
		private const string PINSAFE_MODULE_NAME = "PINsafeOWAFilter";
		private const string PINSAFE_START_COMMENT = "PINsafe settings";
		private const string PINSAFE_END_COMMENT = "End of PINsafe settings";
	    private const string LINE_START_TEXT = "\r\n\t";
		private const string EXCHANGE_VERSION_PATTERN = "14.*";
		private const string PINSAFE_PASSWORD = "PINsafePassword";
		private const string PINSAFE_IMAGE = "PINsafeImage";
		private const string PINSAFE_MESSAGE = "PINsafeMessage";
		private const string PINSAFE_COOKIE_SECRET_CHANGE_INTERVAL = "PINsafeCookieSecretChangeInterval";
		private const string PINSAFE_TARGET_SERVER = "PINsafeTargetServer";
		private const string PINSAFE_TARGET_PATH = "PINsafeTargetPath";
		private const string PINSAFE_EXCLUDED_PATHS = "PINsafeExcludedPaths";
		private const string PINSAFE_LOGON_PATH = "PINsafeLogonPath";
		private const string PINSAFE_LOGOFF_PATH = "PINsafeLogoffPath";
		private const string PINSAFE_AUTH_URL = "PINsafeAuthUrl";
		private const string PINSAFE_IDLE_TIME = "PINsafeIdleTimeSecs";
		private const string ALLOW_NON_PINSAFE_USERS = "AllowNonPINsafeUsers";
		private const string PINSAFE_FILTER_ENABLED = "PINsafeFilterEnabled";
		private const string PINSAFE_INCLUDE_IPS = "PINsafeIncludeIPs";
		private const string PINSAFE_EXCLUDED_IPS = "PINsafeExcludedIPs";
        private const string PINSAFE_IGNORE_DOMAIN_PREFIX = "PINsafeIgnoreDomainPrefix";  
        private const string PINSAFE_IGNORE_DOMAIN_SUFFIX = "PINsafeIgnoreDomainSuffix";
	    private const string PINSAFE_ENABLE_DEBUG = "PINsafeEnableDebug";
	    private const string PINSAFE_DEBUG_LOCATION = "PINsafeDebugLocation";
	    private const string PINSAFE_MULTI_USERNAME = "PINsafeMultiUsername";
	    private const string PINSAFE_AUTO_SHOW = "PINsafeAutoShow";
		private const string PINSAFE_SHOW_PINPAD = "PINsafeUsePinpad";
		private const string CURRENT_OWA_BUILD = "CurrentOWABuild";
		private const string SHOW_CHANGE_PIN = "PINsafeShowChangePIN";
		private const string CHANGE_PIN_REDIRECT = "PINsafeRedirectChangePIN";

        private readonly TimeSpan DEFAULT_CHANGE_INTERVAL = new TimeSpan(2, 0, 0);

	    // Server settings
	    private readonly List<ServerSettings> altServers;
		private bool showPassword;
		private bool showTURing;
		private bool showMessage;
		private long idleTimeSecs;
		private bool allowNonPINsafeUsers;
		private bool filterEnabled;
		private bool includeIPs;
	    private bool ignoreDomainPrefix;
	    private bool ignoreDomainSuffix;
	    private bool enableMultiUsername;
	    private bool autoShow;
		private bool showPinpad;
		private bool showChangePin;
		private bool redirectChangePin;

		public PINsafeSettings(bool autoLoad) {
            altServers = new List<ServerSettings>();
		    if (autoLoad) {
				LoadSettings();
			}
		}

		public PINsafeSettings()
			: this(true) {}

	    public String Server {
            get { return ActiveServer.Server; }
            set { ActiveServer.Server = value; }
	    }

	    public int ServerCount {
            get { return altServers.Count; }
	    }

        public ServerSettings GetServer(int idx) {
            return (idx >= 0 && idx < ServerCount) ? altServers[idx] : null;
        }

	    public List<ServerSettings> Servers {
            get { return altServers; }
	    }

	    public int Port {
            get { return ActiveServer.Port; }
            set { ActiveServer.Port = value; }
	    }

	    public String Context {
            get { return ActiveServer.Context; }
            set { ActiveServer.Context = value; }
	    }

	    public bool Ssl {
            get { return ActiveServer.Ssl; }
            set { ActiveServer.Ssl = value; }
	    }

		public String ProxyServer {
			get { return ActiveServer.ProxyServer; }
			set { ActiveServer.ProxyServer = value; }
		}

		public int ProxyPort {
			get { return ActiveServer.ProxyPort; }
			set { ActiveServer.ProxyPort = value; }
		}

		public String ProxyContext {
			get { return ActiveServer.ProxyContext; }
			set { ActiveServer.ProxyContext = value; }
		}

		public bool ProxySsl {
			get { return ActiveServer.ProxySsl; }
			set { ActiveServer.ProxySsl = value; }
		}

	    public String Secret {
            get { return ActiveServer.Secret; }
            set { ActiveServer.Secret = value; }
	    }

	    public ServerSettings ActiveServer {
            get { return altServers[ActiveServerId]; }
	    }

        public int ActiveServerId { get; set; }

	    public String ServerUrl {
            get { return ActiveServer.Address; }
	    }

		public String ProxyServerUrl {
			get { return ActiveServer.ProxyAddress; }
		}

		public bool ShowPassword {
			get { return showPassword; }
			set { showPassword = value; }
		}

		public bool ShowTURing {
			get { return showTURing; }
			set { showTURing = value; }
		}

		public bool ShowMessage {
			get { return showMessage; }
			set { showMessage = value; }
		}

		public bool AcceptSelfSigned {
            get { return ActiveServer.AcceptSelfSigned; }
            set { ActiveServer.AcceptSelfSigned = value; }
		}

		public bool ShowPinpad {
			get { return showPinpad; }
			set { showPinpad = value; }
		}

		public bool ShowChangePin {
			get { return showChangePin; }
			set { showChangePin = value; }
		}

		public bool RedirectChangePin {
			get { return redirectChangePin; }
			set { redirectChangePin = value; }
		}

		public static string CookieSecret { get; private set; }

        public static string LastCookieSecret { get; private set; }

        public static DateTime CookieLastChanged { get; private set; }

        public TimeSpan CookieSecretChangeInterval { get; set; }

		public string TargetServer { get; set; }

		public string TargetUrl { get; set; }

		public string[] ExcludedPaths { get; set; }

		public string LogonUrl { get; set; }

		public string LogoffUrl { get; set; }

		public string AuthUrl { get; set; }

		public long IdleTimeSecs {
			get { return idleTimeSecs; }
			set { idleTimeSecs = value; }
		}

		public bool AllowNonPINsafeUsers {
			get { return allowNonPINsafeUsers; }
			set { allowNonPINsafeUsers = value; }
		}

		public bool FilterEnabled {
			get { return filterEnabled; }
			set { filterEnabled = value; }
		}

		public bool IncludeIPs {
			get { return includeIPs; }
			set { includeIPs = value; }
		}

        public bool IgnoreDomainPrefix {
            get { return ignoreDomainPrefix; }
            set { ignoreDomainPrefix = value; }
        }

        public bool IgnoreDomainSuffix {
            get { return ignoreDomainSuffix; }
            set { ignoreDomainSuffix = value; }
        }

	    public bool EnableMultipleUsernames {
            get { return enableMultiUsername; }
	    }

	    public bool AutoShow {
	        get { return autoShow; }
            set { autoShow = value; } 
	}

		public IPAddressRange[] ExcludedIPs { get; set; }

		public string CurrentVersion { get; set; }

		public void LoadSettings() {
            altServers.Clear();
            ServerSettings newServer = new ServerSettings();
            newServer.LoadSettings("");
            altServers.Add(newServer);
		    int suffix = 2;
            newServer = new ServerSettings();
		    while (newServer.LoadSettings(suffix.ToString(CultureInfo.InvariantCulture))) {
		        altServers.Add(newServer);
                newServer = new ServerSettings();
		        suffix ++;
		    }
            ActiveServerId = 0;
            String showPw = ConfigurationManager.AppSettings[PINSAFE_PASSWORD];
			String showImg = ConfigurationManager.AppSettings[PINSAFE_IMAGE];
		    String sAutoShow = ConfigurationManager.AppSettings[PINSAFE_AUTO_SHOW];
			String showMsg = ConfigurationManager.AppSettings[PINSAFE_MESSAGE];
			String sShowPinpad = ConfigurationManager.AppSettings[PINSAFE_SHOW_PINPAD];
			String sShowChangePin = ConfigurationManager.AppSettings[SHOW_CHANGE_PIN];
			String sRedirectChangePin = ConfigurationManager.AppSettings[CHANGE_PIN_REDIRECT];
			String sChangeInterval = ConfigurationManager.AppSettings[PINSAFE_COOKIE_SECRET_CHANGE_INTERVAL];
            if (String.IsNullOrEmpty(sChangeInterval)) {
                CookieSecretChangeInterval = DEFAULT_CHANGE_INTERVAL;
            } else {
                int changeHours;
                CookieSecretChangeInterval = int.TryParse(sChangeInterval, out changeHours) ? new TimeSpan(changeHours, 0, 0) : DEFAULT_CHANGE_INTERVAL;
            }
            UpdateSecret();
		    TargetServer = ConfigurationManager.AppSettings[PINSAFE_TARGET_SERVER];
			TargetUrl = TargetServer + ConfigurationManager.AppSettings[PINSAFE_TARGET_PATH];
			string excPathString = ConfigurationManager.AppSettings[PINSAFE_EXCLUDED_PATHS];
			ExcludedPaths = excPathString == null ? new string[0] : excPathString.Split(';');
			LogonUrl = TargetServer + ConfigurationManager.AppSettings[PINSAFE_LOGON_PATH];
			LogoffUrl = TargetServer + ConfigurationManager.AppSettings[PINSAFE_LOGOFF_PATH];
			AuthUrl = ConfigurationManager.AppSettings[PINSAFE_AUTH_URL];
			CurrentVersion = ConfigurationManager.AppSettings[CURRENT_OWA_BUILD];
			String sIdleTimeSecs = ConfigurationManager.AppSettings[PINSAFE_IDLE_TIME];
			String sAllowNonPINsafeUsers = ConfigurationManager.AppSettings[ALLOW_NON_PINSAFE_USERS];
			String sFilterEnabled = ConfigurationManager.AppSettings[PINSAFE_FILTER_ENABLED];
            String sIgnoreDomainPrefix = ConfigurationManager.AppSettings[PINSAFE_IGNORE_DOMAIN_PREFIX];
            String sIgnoreDomainSuffix = ConfigurationManager.AppSettings[PINSAFE_IGNORE_DOMAIN_SUFFIX];
            if (!Boolean.TryParse(showPw, out showPassword)) {
				showPassword = true;
			}
			if (!Boolean.TryParse(showImg, out showTURing)) {
				showTURing = true;
			}
            if (!Boolean.TryParse(sAutoShow, out autoShow)) {
                autoShow = true;
            }
			if (!Boolean.TryParse(showMsg, out showMessage)) {
				showMessage = false;
			}
			if (!Boolean.TryParse(sShowPinpad, out showPinpad)) {
				showPinpad = false;
			}
			if (!Boolean.TryParse(sShowChangePin, out showChangePin)) {
				showChangePin = true;
			}
			if (!Boolean.TryParse(sRedirectChangePin, out redirectChangePin)) {
				redirectChangePin = true;
			}
			if (!Int64.TryParse(sIdleTimeSecs, out idleTimeSecs)) {
				idleTimeSecs = 300; // Default 5 minutes.
			}
			if (!Boolean.TryParse(sAllowNonPINsafeUsers, out allowNonPINsafeUsers)) {
				allowNonPINsafeUsers = false;
			}
			if (!Boolean.TryParse(sFilterEnabled, out filterEnabled)) {
				filterEnabled = true;
			}
            if (!Boolean.TryParse(sIgnoreDomainPrefix, out ignoreDomainPrefix)) {
                ignoreDomainPrefix = true;
            }
            if (!Boolean.TryParse(sIgnoreDomainSuffix, out ignoreDomainSuffix)) {
                ignoreDomainSuffix = true;
            }
            string sIncludeIPs = ConfigurationManager.AppSettings[PINSAFE_INCLUDE_IPS];
			if (String.IsNullOrEmpty(sIncludeIPs)) {
				IncludeIPs = false;
			}
			else {
				if (!Boolean.TryParse(sIncludeIPs, out includeIPs)) {
					IncludeIPs = false;
				}
			}
			string excIPs = ConfigurationManager.AppSettings[PINSAFE_EXCLUDED_IPS];
			if (String.IsNullOrEmpty(excIPs)) {
				ExcludedIPs = new IPAddressRange[0];
			}
			else {
				string[] aExcIPs = excIPs.Split(';');
				foreach (IPAddressRange range in aExcIPs.Select(t => new IPAddressRange(t)).Where(range => range.IsValid)) {
					AddExcludedIP(range);
				}
			}
		    string sEnableDebug = ConfigurationManager.AppSettings[PINSAFE_ENABLE_DEBUG];
		    bool enableDebug = false;
            if (!String.IsNullOrEmpty(sEnableDebug)) {
                if (!Boolean.TryParse(sEnableDebug, out enableDebug)) {
                    enableDebug = false;
                }
            }
		    string debugLocation = ConfigurationManager.AppSettings[PINSAFE_DEBUG_LOCATION];
            Debug.Initialize(enableDebug, debugLocation);
            string sMultiUser = ConfigurationManager.AppSettings[PINSAFE_MULTI_USERNAME];
            enableMultiUsername = false;
            if (!String.IsNullOrEmpty(sMultiUser)) {
                if (!Boolean.TryParse(sMultiUser, out enableMultiUsername)) {
                    enableMultiUsername = false;
                }
            }
        }

		private static Dictionary<string, string> ReadSettings(XmlElement parent) {
			return parent.GetElementsByTagName("add").Cast<XmlElement>().ToDictionary(elem => elem.GetAttribute("key"),
			                                                                          elem => elem.GetAttribute("value"));
		}

		internal static string GetSetting(Dictionary<string, string> settings, string name) {
			return (settings.ContainsKey(name)) ? settings[name] : "";
		}

		internal static bool GetBoolSetting(Dictionary<string, string> settings, string name, bool defValue) {
			if (settings.ContainsKey(name)) {
				string sValue = settings[name];
				bool value;
				if (Boolean.TryParse(sValue, out value)) {
					return value;
				}
			}
			return defValue;
		}

		internal static int GetIntSetting(Dictionary<string, string> settings, string name, int defValue) {
			if (settings.ContainsKey(name)) {
				string sValue = settings[name];
				int value;
				if (Int32.TryParse(sValue, out value)) {
					return value;
				}
			}
			return defValue;
		}

		public bool LoadSettings(string location) {
			if (File.Exists(location)) {
				using (StreamReader reader = new StreamReader(location)) {
                    try {
                        XmlDocument configDoc = new XmlDocument();
                        configDoc.Load(reader);
                        if (configDoc.DocumentElement != null) {
                            XmlElement settingsElem = configDoc.DocumentElement["appSettings"];
                            if (settingsElem != null) {
                                Dictionary<string, string> settings = ReadSettings(settingsElem);
                                // Just check that we have at least one PINsafe entry.
                                if (settings.ContainsKey(PINSAFE_TARGET_SERVER)) {
                                    altServers.Clear();
                                    ServerSettings newServer = new ServerSettings();
                                    newServer.LoadSettings(settings, "");
                                    altServers.Add(newServer);
                                    int suffix = 2;
                                    newServer = new ServerSettings();
                                    while (newServer.LoadSettings(settings,
                                                                  suffix.ToString(CultureInfo.InvariantCulture))) {
                                        altServers.Add(newServer);
                                        newServer = new ServerSettings();
                                        suffix++;
                                    }
                                    ActiveServerId = 0;
                                    showPassword = GetBoolSetting(settings, PINSAFE_PASSWORD, true);
                                    showTURing = GetBoolSetting(settings, PINSAFE_IMAGE, true);
                                    showMessage = GetBoolSetting(settings, PINSAFE_MESSAGE, false);
                                    autoShow = GetBoolSetting(settings, PINSAFE_AUTO_SHOW, true);
	                                showPinpad = GetBoolSetting(settings, PINSAFE_SHOW_PINPAD, false);
	                                showChangePin = GetBoolSetting(settings, SHOW_CHANGE_PIN, true);
	                                redirectChangePin = GetBoolSetting(settings, CHANGE_PIN_REDIRECT, true);
                                    int changeHours = GetIntSetting(settings, PINSAFE_COOKIE_SECRET_CHANGE_INTERVAL, 0);
                                    CookieSecretChangeInterval = (changeHours == 0)
                                                                     ? new TimeSpan(changeHours, 0, 0)
                                                                     : DEFAULT_CHANGE_INTERVAL;
                                    UpdateSecret();
                                    TargetServer = GetSetting(settings, PINSAFE_TARGET_SERVER);
                                    TargetUrl = TargetServer + GetSetting(settings, PINSAFE_TARGET_PATH);
                                    string sExcPaths = GetSetting(settings, PINSAFE_EXCLUDED_PATHS);
                                    ExcludedPaths = (sExcPaths == null) ? new string[0] : sExcPaths.Split(';');
                                    if (ExcludedPaths.Length == 1 && "".Equals(ExcludedPaths[0])) {
                                        ExcludedPaths = new string[0];
                                    }
                                    LogonUrl = TargetServer + GetSetting(settings, PINSAFE_LOGON_PATH);
                                    LogoffUrl = TargetServer + GetSetting(settings, PINSAFE_LOGOFF_PATH);
                                    AuthUrl = GetSetting(settings, PINSAFE_AUTH_URL);
                                    idleTimeSecs = GetIntSetting(settings, PINSAFE_IDLE_TIME, 300);
                                    allowNonPINsafeUsers = GetBoolSetting(settings, ALLOW_NON_PINSAFE_USERS, false);
                                    filterEnabled = GetBoolSetting(settings, PINSAFE_FILTER_ENABLED, true);
                                    includeIPs = GetBoolSetting(settings, PINSAFE_INCLUDE_IPS, false);
                                    ignoreDomainPrefix = GetBoolSetting(settings, PINSAFE_IGNORE_DOMAIN_PREFIX, true);
                                    ignoreDomainSuffix = GetBoolSetting(settings, PINSAFE_IGNORE_DOMAIN_SUFFIX, false);
                                    string sExcIPs = GetSetting(settings, PINSAFE_EXCLUDED_IPS);
                                    if (String.IsNullOrEmpty(sExcIPs)) {
                                        ExcludedIPs = new IPAddressRange[0];
                                    } else {
                                        string[] aExcIPs = sExcIPs.Split(';');
                                        foreach (
                                            IPAddressRange range in
                                                aExcIPs.Select(t => new IPAddressRange(t)).Where(range => range.IsValid)
                                            ) {
                                            AddExcludedIP(range);
                                        }
                                    }
	                                CurrentVersion = GetSetting(settings, CURRENT_OWA_BUILD);
                                    bool enableDebug = GetBoolSetting(settings, PINSAFE_ENABLE_DEBUG, false);
                                    string debugLocation = GetSetting(settings, PINSAFE_DEBUG_LOCATION);
                                    Debug.Initialize(enableDebug, debugLocation);
                                    enableMultiUsername = GetBoolSetting(settings, PINSAFE_MULTI_USERNAME, false);
                                    return true;
                                }
                            }
                        }
                    } catch (XmlException) {
                        // Not a valid XML document
                        // Use defaults
                    }
				}
			}
            SetDefaults();
            return false;
        }

		private static void WriteSettings(XmlDocument configDoc, XmlElement parent, Dictionary<string, string> settings) {
			XmlNodeList nodes = parent.ChildNodes;
			XmlNode endComment = null;
			for (int idx = 0; idx < nodes.Count; idx++) {
				XmlElement element = nodes[idx] as XmlElement;
				if (element != null) {
					XmlElement elem = element;
					if (elem.Name.Equals("add")) {
						string name = elem.GetAttribute("key");
						if (settings.ContainsKey(name)) {
							elem.SetAttribute("value", settings[name]);
							settings.Remove(name);
						}
						else {
							parent.RemoveChild(elem);
						}
					}
				}
				else if (nodes[idx] is XmlComment) {
					string comment = nodes[idx].InnerText;
					if (comment.ToLower().Contains(PINSAFE_END_COMMENT.ToLower())) {
						endComment = nodes[idx];
					}
				}
			}
			if (endComment == null) {
				// Create comments to bracket the PINsafe entries.
			    XmlText startText = configDoc.CreateTextNode(LINE_START_TEXT);
			    parent.AppendChild(startText);
				XmlComment startComment = configDoc.CreateComment(PINSAFE_START_COMMENT);
				parent.AppendChild(startComment);
                startText = configDoc.CreateTextNode(LINE_START_TEXT);
                parent.AppendChild(startText);
                endComment = configDoc.CreateComment(PINSAFE_END_COMMENT);
				parent.AppendChild(endComment);
                startText = configDoc.CreateTextNode(LINE_START_TEXT);
                parent.AppendChild(startText);
            }
			foreach (KeyValuePair<string, string> entry in settings) {
				XmlElement elem = configDoc.CreateElement("add");
				elem.SetAttribute("key", entry.Key);
				elem.SetAttribute("value", entry.Value);
				parent.InsertBefore(elem, endComment);
                XmlText startText = configDoc.CreateTextNode(LINE_START_TEXT);
                parent.InsertBefore(startText, endComment);
            }
		}

		private static void LocateModulesElement(XmlElement configElem, out XmlElement baseElem,
		                                         out string systemWebElem, out string modulesElem) {
			systemWebElem = null;
			modulesElem = null;
			baseElem = configElem;

			XmlNodeList locationElems = configElem.GetElementsByTagName("location");
			foreach (XmlElement locationElem in locationElems) {
				string pathAttr = locationElem.GetAttribute("path");
			    if (!String.IsNullOrEmpty(pathAttr)) {
			        continue;
			    }
			    // Check if this element contains an Http modules section.
			    XmlElement sysWebElem = locationElem["system.web"];
			    if (sysWebElem != null) {
			        if (sysWebElem["httpModules"] != null) {
			            baseElem = locationElem;
			            systemWebElem = "system.web";
			            modulesElem = "httpModules";
			            break;
			        }
			    }
			    sysWebElem = locationElem["system.webServer"];
			    if (sysWebElem == null) {
			        continue;
			    }
			    if (sysWebElem["modules"] == null) {
			        continue;
			    }
			    baseElem = locationElem;
			    systemWebElem = "system.webServer";
			    modulesElem = "modules";
			    break;
			}
			if (baseElem == configElem) {
				// Check what version of IIS we are using.
				Int32 iisVersion;
				using (RegistryKey iisKey = Registry.LocalMachine.OpenSubKey(@"Software\Microsoft\InetStp")) {
					if (iisKey == null) {
						baseElem = null;
						return; // Unable to determine IIS version.
					}
					iisVersion = (Int32) iisKey.GetValue("MajorVersion");
					if (iisVersion < 5) {
						baseElem = null;
						return;
					}
				}
				// We haven't found the modules section in any location element.
				if (iisVersion == 7) {
					// Check if the system.web/httpModules element already exists.
					XmlElement sysWebElem = configElem["system.web"];
					if (sysWebElem != null && sysWebElem["httpModules"] != null) {
						systemWebElem = "system.web";
						modulesElem = "httpModules";
					}
					else {
						sysWebElem = configElem["system.webServer"];
						if (sysWebElem != null && sysWebElem["modules"] != null) {

							systemWebElem = "system.webServer";
							modulesElem = "modules";
						}
					}
				}
				else {
					systemWebElem = "system.web";
					modulesElem = "httpModules";
				}
			}
			if (systemWebElem == null) {
				systemWebElem = "system.webServer";
				modulesElem = "modules";
			}
		}

		public bool SaveSettings(string location) {
			if (File.Exists(location)) {
				XmlDocument configDoc = new XmlDocument {PreserveWhitespace = true};
				using (StreamReader reader = new StreamReader(location)) {
					configDoc.Load(reader);
				}
				XmlElement configElem = configDoc.DocumentElement;
				if (configElem == null) {
					return false;
				}
				XmlElement settingsElem = configElem["appSettings"];
				if (settingsElem == null) {
					settingsElem = configDoc.CreateElement("appSettings");
					configElem.AppendChild(settingsElem);
				}

				// Check whether the filter (HttpModule) has been entered into the web.config.
				string systemWebElem;
				string modulesElem;
				XmlElement baseElem;
				LocateModulesElement(configElem, out baseElem, out systemWebElem, out modulesElem);
				if (baseElem == null) {
					return false;
				}
				XmlElement webElem = baseElem[systemWebElem];
				if (webElem == null) {
					webElem = configDoc.CreateElement(systemWebElem);
					baseElem.AppendChild(webElem);
				}
				XmlElement moduleElem = webElem[modulesElem];
				if (moduleElem == null) {
					moduleElem = configDoc.CreateElement(modulesElem);
					webElem.AppendChild(moduleElem);
				}
				XmlNodeList modules = moduleElem.GetElementsByTagName("add");
				bool moduleExists = (from XmlElement elem in modules select elem.GetAttribute("type")).Any(type => type != null && type.Equals(PINSAFE_MODULE_TYPE));
				if (!moduleExists) {
					XmlElement newModule = configDoc.CreateElement("add");
					newModule.SetAttribute("type", PINSAFE_MODULE_TYPE);
					newModule.SetAttribute("name", PINSAFE_MODULE_NAME);
					moduleElem.AppendChild(newModule);
				}
				// BUGFIX: Get around the embedded newlines in the Exchange assemblies by removing them.
				XmlElement compElem = webElem["compilation"];
				if (compElem != null) {
					XmlElement assembliesElem = compElem["assemblies"];
					if (assembliesElem != null) {
						XmlNodeList assemblies = assembliesElem.GetElementsByTagName("add");
						foreach (XmlElement assemblyElem in assemblies) {
							string assemblyAttr = assemblyElem.GetAttribute("assembly");
							string[] split = assemblyAttr.Split(',');
							StringBuilder newAttr = new StringBuilder(split[0].Trim());
							for (int idx = 1; idx < split.Length; idx++) {
								newAttr.Append(", ");
								newAttr.Append(split[idx].Trim());
							}
							assemblyElem.SetAttribute("assembly", newAttr.ToString());
						}
					}
				}

				Dictionary<string, string> settings = ReadSettings(settingsElem);
				// Write all the settings to the XML file.
                altServers[0].SaveSettings(settings, "");
                for (int idx = 2; idx <= altServers.Count; idx++) {
                    altServers[idx-1].SaveSettings(settings, idx.ToString(CultureInfo.InvariantCulture));
                }
                settings[PINSAFE_PASSWORD] = showPassword.ToString(CultureInfo.InvariantCulture);
				settings[PINSAFE_IMAGE] = showTURing.ToString(CultureInfo.InvariantCulture);
				settings[PINSAFE_MESSAGE] = showMessage.ToString(CultureInfo.InvariantCulture);
			    settings[PINSAFE_AUTO_SHOW] = autoShow.ToString(CultureInfo.InvariantCulture);
				settings[PINSAFE_SHOW_PINPAD] = showPinpad.ToString(CultureInfo.InvariantCulture);
				settings[SHOW_CHANGE_PIN] = showChangePin.ToString(CultureInfo.InvariantCulture);
				settings[CHANGE_PIN_REDIRECT] = redirectChangePin.ToString(CultureInfo.InvariantCulture);
				settings[PINSAFE_COOKIE_SECRET_CHANGE_INTERVAL] = CookieSecretChangeInterval.Hours.ToString(CultureInfo.InvariantCulture);
				settings[PINSAFE_TARGET_SERVER] = TargetServer;
				settings[PINSAFE_TARGET_PATH] = TargetUrl.Substring(TargetServer.Length);
				if (ExcludedPaths.Length > 0) {
					StringBuilder exPathString = new StringBuilder(ExcludedPaths[0]);
					for (int idx = 1; idx < ExcludedPaths.Length; idx++) {
						exPathString.Append(";");
						exPathString.Append(ExcludedPaths[idx]);
					}
					settings[PINSAFE_EXCLUDED_PATHS] = exPathString.ToString();
				} else {
					settings[PINSAFE_EXCLUDED_PATHS] = "";
				}
				settings[PINSAFE_LOGON_PATH] = LogonUrl.Substring(TargetServer.Length);
				settings[PINSAFE_LOGOFF_PATH] = LogoffUrl.Substring(TargetServer.Length);
				settings[PINSAFE_AUTH_URL] = AuthUrl;
				settings[PINSAFE_IDLE_TIME] = IdleTimeSecs.ToString(CultureInfo.InvariantCulture);
				settings[ALLOW_NON_PINSAFE_USERS] = AllowNonPINsafeUsers.ToString(CultureInfo.InvariantCulture);
				settings[PINSAFE_FILTER_ENABLED] = FilterEnabled.ToString(CultureInfo.InvariantCulture);
				settings[PINSAFE_INCLUDE_IPS] = IncludeIPs.ToString(CultureInfo.InvariantCulture);
                settings[PINSAFE_IGNORE_DOMAIN_PREFIX] = IgnoreDomainPrefix.ToString(CultureInfo.InvariantCulture);
                settings[PINSAFE_IGNORE_DOMAIN_SUFFIX] = IgnoreDomainSuffix.ToString(CultureInfo.InvariantCulture);
				settings[CURRENT_OWA_BUILD] = CurrentVersion;
                if (ExcludedIPs.Length > 0) {
					StringBuilder excIPString = new StringBuilder(ExcludedIPs[0].ToString());
					for (int idx = 1; idx < ExcludedIPs.Length; idx++) {
						excIPString.Append(";");
						excIPString.Append(ExcludedIPs[idx]);
					}
					settings[PINSAFE_EXCLUDED_IPS] = excIPString.ToString();
				} else {
					settings[PINSAFE_EXCLUDED_IPS] = "";
				}
				WriteSettings(configDoc, settingsElem, settings);
				// Make a backup.
				string backupName = location + ".bak";
				if (File.Exists(backupName)) {
					File.Delete(backupName);
				}
				File.Move(location, backupName);
				using (StreamWriter writer = new StreamWriter(location)) {
					//XmlWriterSettings xSettings = new XmlWriterSettings {Indent = true, IndentChars = "\t"};
					XmlWriter w = XmlWriter.Create(writer);
					try {
						configDoc.Save(w);
					} catch {
						return false;
					}
				}
			}
			return true;
		}

        public void SetDefaults() {
            const string OWA_LOCATION = @"C:\Program Files\Microsoft\Exchange Server\V14";
			const string HOSTNAME = "localhost";
			// Configure sensible default values.
            altServers.Clear();
            altServers.Add(new ServerSettings());
		    ActiveServerId = 0;
			Server = "";
			Port = 8080;
			Context = "pinsafe";
			Secret = "";
			Ssl = true;
			ProxyServer = "";
			ProxyPort = 8443;
			ProxyContext = "proxy";
			ProxySsl = true;
			showPassword = false;
			showTURing = true;
			showMessage = false;
		    autoShow = true;
			showPinpad = false;
	        showChangePin = true;
	        redirectChangePin = true;
			AcceptSelfSigned = false;
			CookieSecret = "";
			TargetServer = "https://" + HOSTNAME;
			TargetUrl = TargetServer + DEFAULT_OWA_URL;
			ExcludedPaths = new[] {DEFAULT_AUTH_URL, DEFAULT_AUTH_FOLDER};
            // Get the current Exchange version
            CheckCurrentOWABuild(OWA_LOCATION);
			LogonUrl = TargetServer + DEFAULT_AUTH_FOLDER + "Logon.aspx";
            LogoffUrl = TargetServer + DEFAULT_AUTH_FOLDER + "Logoff.aspx";
			AuthUrl = TargetServer + DEFAULT_AUTH_URL;
			idleTimeSecs = 300;
			allowNonPINsafeUsers = false;
			filterEnabled = true;
			IncludeIPs = false;
			ExcludedIPs = new IPAddressRange[0];
            ignoreDomainPrefix = true;
            ignoreDomainSuffix = false;
            Debug.Initialize(true);
		    enableMultiUsername = false;
		}

		// Make sure the current OWA build folder is in the list of allowed paths.
		public void CheckCurrentOWABuild(string owaRoot) {
		    if (!Directory.Exists(owaRoot)) {
		        return;
		    }
		    string[] verFolders = Directory.GetDirectories(owaRoot, EXCHANGE_VERSION_PATTERN);
		    if (verFolders.Length <= 0) {
		        return;
		    }
		    Array.Sort(verFolders);
		    CurrentVersion = verFolders[verFolders.Length - 1];
		    CurrentVersion = CurrentVersion.Substring(CurrentVersion.LastIndexOf('\\') + 1);
		    AddExcludedPath(DEFAULT_OWA_URL + CurrentVersion + "/");
		}

		private void AddExcludedPath(string newPath) {
			if (String.IsNullOrEmpty(newPath)) {
				return;
			}
			string[] oldPaths = ExcludedPaths;
			if (oldPaths.Any(oldPath => String.Compare(oldPath, newPath, StringComparison.OrdinalIgnoreCase) == 0)) {
				return;
			}
			// OK, the path needs to be added.
			string[] newPaths = new string[oldPaths.Length + 1];
			for (int idx = 0; idx < oldPaths.Length; idx++) {
				newPaths[idx] = oldPaths[idx];
			}
			newPaths[oldPaths.Length] = newPath;
			ExcludedPaths = newPaths;
		}

		private void AddExcludedIP(IPAddressRange newIP) {
			if (newIP == null || !newIP.IsValid) {
				return;
			}
			IPAddressRange[] oldIPs = ExcludedIPs;
			if (oldIPs != null && oldIPs.Any(oldPath => IPAddressRange.Equals(oldPath, newIP))) {
				return;
			}
			// OK, the path needs to be added.
			IPAddressRange[] newIPs;
			if (oldIPs == null) {
				newIPs = new IPAddressRange[1];
				newIPs[0] = newIP;
			} else {
				newIPs = new IPAddressRange[oldIPs.Length + 1];
				for (int idx = 0; idx < oldIPs.Length; idx++) {
					newIPs[idx] = oldIPs[idx];
				}
				newIPs[oldIPs.Length] = newIP;
			}
			ExcludedIPs = newIPs;
		}

		public void RemoveSettings(string configFile) {
			if (!File.Exists(configFile))
				return;
			XmlDocument configDoc = new XmlDocument {PreserveWhitespace = true};
			using (StreamReader reader = new StreamReader(configFile)) {
				using (XmlTextReader xReader = new XmlTextReader(reader)) {
					// We don't want to modify any of the document except for the bits that are relevant.
					configDoc.Load(xReader);
				}
			}
			XmlElement configElem = configDoc.DocumentElement;
			if (configElem == null) {
				return;
			}

			XmlElement baseElem;
			string systemWebElem;
			string modulesElem;
			LocateModulesElement(configElem, out baseElem, out systemWebElem, out modulesElem);
			if (baseElem == null) {
				return;
			}

			// Check whether the filter (HttpModule) has been entered into the web.config.
			XmlElement webElem = baseElem[systemWebElem];
			if (webElem != null) {
				XmlElement moduleElem = webElem[modulesElem];
				if (moduleElem != null) {
					XmlNodeList modules = moduleElem.GetElementsByTagName("add");
					foreach (XmlElement elem in modules) {
						string type = elem.GetAttribute("type");
						if (!type.Equals(PINSAFE_MODULE_TYPE))
							continue;
						moduleElem.RemoveChild(elem);
						break;
					}
				}
			}
			XmlElement settingsElem = configElem["appSettings"];
			if (settingsElem != null) {
				bool started = false;
				XmlNode node = settingsElem.FirstChild;
				while (node != null) {
					if (node is XmlComment && node.InnerText.ToLower().Contains(PINSAFE_START_COMMENT.ToLower())) {
						started = true;
						XmlNode oldNode = node;
						node = oldNode.NextSibling;
						settingsElem.RemoveChild(oldNode);
					}
					else if (started && node is XmlElement && node.Name == "add") {
						XmlNode oldNode = node;
						node = oldNode.NextSibling;
						settingsElem.RemoveChild(oldNode);
					}
					else if (started && node is XmlComment &&
					         node.InnerText.ToLower().Contains(PINSAFE_END_COMMENT.ToLower())) {
						settingsElem.RemoveChild(node);
						node = null;
					}
					else {
						node = node.NextSibling;
					}
				}
			}
			// Make a backup.
			string backupName = configFile + ".bak";
			if (File.Exists(backupName)) {
				File.Delete(backupName);
			}
			File.Move(configFile, backupName);
			using (StreamWriter writer = new StreamWriter(configFile)) {
				configDoc.Save(writer);
			}
		}

        private void UpdateSecret() {
            if (String.IsNullOrEmpty(CookieSecret)) {
                // Secret has not yet been set.
                CookieSecret = GenerateRandomSecret();
                CookieLastChanged = DateTime.Now;
                LastCookieSecret = null;
            } else if (CookieLastChanged < DateTime.Now.Add(CookieSecretChangeInterval)) {
                // Secret has expired.
                LastCookieSecret = CookieSecret;
                CookieSecret = GenerateRandomSecret();
                CookieLastChanged = DateTime.Now;
            }
        }

	    private static String GenerateRandomSecret(int length = 0x30) {
	        char[] cookie = new char[length];
	        Random rnd = new Random();
	        for (int idx = 0; idx < length; idx++) {
	            cookie[idx] = (char)(rnd.Next(0x20, 0x60));
	        }
	        return new string(cookie);
	    }
	}
}