using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Reflection;
using System.Windows.Forms;
using com.swivelsecure.pinsafe;

namespace swivelsecure.pinsafe.config {
    public partial class ConfigForm : Form {

        private string owaLocation;
        private bool settingsChanged;
        private bool dirty;
        private readonly PINsafeSettings settings;
        private int activeServerId;
        private ServersForm serverForm;

        public string OwaLocation {
            get { return owaLocation; }
            set {
                owaLocation = value;
                configFile = owaLocation + "\\web.config";
                settings.LoadSettings(configFile);
            }
        }

        public ConfigForm() {
            InitializeComponent();
            settingsChanged = false;
            settings = new PINsafeSettings(false);
            serverForm = null;
            dirty = false;
        }

        private void SaveSettings() {
            if (!CheckValues())
                return;

            if (!settings.SaveSettings(configFile)) {
                return;
            }
            settingsChanged = true;
            dirty = false;
        }

        private void BtnApplyClick(object sender, EventArgs e) {
            SaveSettings();
        }

        private void BtnCloseClick(object sender, EventArgs e) {
            SaveSettings();
            DialogResult = DialogResult.OK;
            Close();
        }

        private void BtnCancelClick(object sender, EventArgs e) {
            if (!AllowExit()) return;
            DialogResult = DialogResult.Cancel;
            Close();
        }

        private bool AllowExit() {
            if (CheckValues()) {
                if (
                    MessageBox.Show("You have not saved your changes. Exit and lose unsaved changes?", "Cancel Changes",
                                    MessageBoxButtons.YesNo) == DialogResult.No) {
                    return false;
                }
            } else if (settingsChanged) {
                if (MessageBox.Show(
                    "Some configuration settings have been changed. You must restart IIS to implement these changes. Restart Now?",
                    "Restart IIS", MessageBoxButtons.YesNo) == DialogResult.Yes) {
                    ResetIIS();
                }
                settingsChanged = false;
            }
            return true;
        }

        private void ConfigFormFormClosing(object sender, FormClosingEventArgs e) {
            if (!AllowExit()) {
                e.Cancel = true;
            }
        }

        private void ConfigFormShow(object sender, EventArgs e) {
            SetValues();
			Assembly asm = Assembly.GetExecutingAssembly();
			Version ver = asm.GetName().Version;
			String verInfo = "Version " + ver;
			object[] oCopyright = asm.GetCustomAttributes(typeof(AssemblyCopyrightAttribute), false);
			if (oCopyright.Length > 0) {
				verInfo += " " + ((AssemblyCopyrightAttribute)oCopyright[0]).Copyright;
			}
			lblAbout.Text = verInfo;
		}

        private void SetValues() {
            settings.LoadSettings(configFile);

            // PINsafe tab
            activeServerId = 0;
            UpdateServer();

            // OWA tab
            string path = settings.TargetServer;
            txtOWAServer.Text = path ?? "";
            int serverLen = (path == null) ? 0 : path.Length;
            path = settings.TargetUrl;
            txtOWAPath.Text = (path == null) ? "" : path.Substring(serverLen);
            path = settings.LogonUrl;
            txtLogonPath.Text = (path == null) ? "" : path.Substring(serverLen);
            path = settings.LogoffUrl;
            txtLogoffPath.Text = (path == null) ? "" : path.Substring(serverLen);
            txtAuthUrl.Text = settings.AuthUrl;

            // Authentication tab
            txtCookieSecretChangeInterval.Text = settings.CookieSecretChangeInterval.Hours.ToString(CultureInfo.InvariantCulture);
            txtIdleTime.Text = settings.IdleTimeSecs.ToString(CultureInfo.InvariantCulture);
            chkAllowNonPINsafe.Checked = settings.AllowNonPINsafeUsers;
            chkFilterEnabled.Checked = settings.FilterEnabled;
            chkIgnoreDomainPrefix.Checked = settings.IgnoreDomainPrefix;
            chkIgnoreDomainSuffix.Checked = settings.IgnoreDomainSuffix;
            chkShowTuring.Checked = settings.ShowTURing;
            chkShowMessage.Checked = settings.ShowMessage;
	        chkShowPinpad.Checked = settings.ShowPinpad;
            chkAutoGenerate.Checked = settings.AutoShow;
	        chkShowChangePIN.Checked = settings.ShowChangePin;
	        chkForceChangePIN.Checked = settings.RedirectChangePin;

            // Excluded tab
            settings.CheckCurrentOWABuild(owaLocation);
            txtExcluded.Lines = settings.ExcludedPaths;
            drpIncludeIPs.SelectedIndex = settings.IncludeIPs ? 1 : 0;
            string[] lines = new string[settings.ExcludedIPs.Length];
            for (int idx = 0; idx < lines.Length; idx++) {
                lines[idx] = settings.ExcludedIPs[idx].ToString();
            }
            txtExcludedIPs.Lines = lines;
        }

        private bool CheckServer() {
            bool changed = false;
            ServerSettings server = settings.GetServer(activeServerId);
            if (String.CompareOrdinal(txtServer.Text, server.Server) != 0) {
                server.Server = txtServer.Text;
                changed = true;
            }
            int newPort;
            if (int.TryParse(txtPort.Text, out newPort)) {
                if (newPort != server.Port) {
                    server.Port = newPort;
                    changed = true;
                }
            }

            if (String.CompareOrdinal(txtContext.Text, server.Context) != 0) {
                server.Context = txtContext.Text;
                changed = true;
            }

            if (String.CompareOrdinal(txtSecret.Text, server.Secret) != 0) {
                if (String.CompareOrdinal(txtSecret.Text, txtSecret2.Text) == 0) {
                    server.Secret = txtSecret.Text;
                    changed = true;
                } else {
                    MessageBox.Show("Secrets do not match. Not changed", "Error", MessageBoxButtons.OK,
                                    MessageBoxIcon.Error);
                }
            }

            if (chkSecure.Checked != server.Ssl) {
                server.Ssl = chkSecure.Checked;
                changed = true;
            }

            if (chkSelfSigned.Checked != server.AcceptSelfSigned) {
                server.AcceptSelfSigned = chkSelfSigned.Checked;
                changed = true;
            }

			if (String.CompareOrdinal(txtProxyServer.Text, server.ProxyServer) != 0) {
				server.ProxyServer = txtProxyServer.Text;
				changed = true;
			}
			if (int.TryParse(txtProxyPort.Text, out newPort)) {
				if (newPort != server.ProxyPort) {
					server.ProxyPort = newPort;
					changed = true;
				}
			}

			if (String.CompareOrdinal(txtProxyContext.Text, server.ProxyContext) != 0) {
				server.ProxyContext = txtProxyContext.Text;
				changed = true;
			}
			if (chkProxySecure.Checked != server.ProxySsl) {
				server.ProxySsl = chkProxySecure.Checked;
				changed = true;
			}

			return changed;
        }

        private bool CheckValues() {
            // Check all values against current settings, and update if necessary.
            bool changed = dirty;

            // PINsafe tab
            changed |= CheckServer();

            // OWA tab
            if (String.CompareOrdinal(txtOWAServer.Text, settings.TargetServer) != 0) {
                settings.TargetServer = txtOWAServer.Text;
                changed = true;
                settings.TargetUrl = settings.TargetServer + txtOWAPath.Text;
                settings.LogonUrl = settings.TargetServer + txtLogonPath.Text;
                settings.LogoffUrl = settings.TargetServer + txtLogoffPath.Text;
            } else {
                int serverLen = settings.TargetServer.Length;
                if (string.Compare(txtOWAPath.Text, 0, settings.TargetUrl, serverLen, 100) != 0) {
                    settings.TargetUrl = settings.TargetServer + txtOWAPath.Text;
                    changed = true;
                }
                if (string.Compare(txtLogonPath.Text, 0, settings.LogonUrl, serverLen, 100) != 0) {
                    settings.LogonUrl = settings.TargetServer + txtLogonPath.Text;
                    changed = true;
                }
                if (string.Compare(txtLogoffPath.Text, 0, settings.LogoffUrl, serverLen, 100) != 0) {
                    settings.LogoffUrl = settings.TargetServer + txtLogoffPath.Text;
                    changed = true;
                }
            }
            if (String.CompareOrdinal(txtAuthUrl.Text, settings.AuthUrl) != 0) {
                settings.AuthUrl = txtAuthUrl.Text;
                changed = true;
            }

            // Authentication tab
            int oldCookieSecretChangeInterval = settings.CookieSecretChangeInterval.Hours;
            int changeHours;
            if (int.TryParse(txtCookieSecretChangeInterval.Text, out changeHours)) {
                if (changeHours != oldCookieSecretChangeInterval) {
                    settings.CookieSecretChangeInterval = new TimeSpan(changeHours, 0, 0);
                    changed = true;
                }
            } else {
                // Invalid change interval: put the old one back.
                txtCookieSecretChangeInterval.Text = oldCookieSecretChangeInterval.ToString(CultureInfo.InvariantCulture);
            }

            int idleTime;
            if (int.TryParse(txtIdleTime.Text, out idleTime)) {
                if (settings.IdleTimeSecs != idleTime) {
                    settings.IdleTimeSecs = idleTime;
                    changed = true;
                }
            }

            if (chkAllowNonPINsafe.Checked != settings.AllowNonPINsafeUsers) {
                settings.AllowNonPINsafeUsers = chkAllowNonPINsafe.Checked;
                changed = true;
            }

            if (chkFilterEnabled.Checked != settings.FilterEnabled) {
                settings.FilterEnabled = chkFilterEnabled.Checked;
                changed = true;
            }

            if (chkIgnoreDomainPrefix.Checked != settings.IgnoreDomainPrefix) {
                settings.IgnoreDomainPrefix = chkIgnoreDomainPrefix.Checked;
                changed = true;
            }

            if (chkIgnoreDomainSuffix.Checked != settings.IgnoreDomainSuffix) {
                settings.IgnoreDomainSuffix = chkIgnoreDomainSuffix.Checked;
                changed = true;
            }

            if (chkShowTuring.Checked != settings.ShowTURing) {
                settings.ShowTURing = chkShowTuring.Checked;
                changed = true;
            }

            if (chkShowMessage.Checked != settings.ShowMessage) {
                settings.ShowMessage = chkShowMessage.Checked;
                changed = true;
            }

			if (chkShowPinpad.Checked != settings.ShowPinpad) {
				settings.ShowPinpad = chkShowPinpad.Checked;
				changed = true;
			}

            if (chkAutoGenerate.Checked != settings.AutoShow) {
                settings.AutoShow = chkAutoGenerate.Checked;
                changed = true;
            }

			if (chkShowChangePIN.Checked != settings.ShowChangePin) {
				settings.ShowChangePin = chkShowChangePIN.Checked;
				changed = true;
			}

			if (chkForceChangePIN.Checked != settings.RedirectChangePin) {
				settings.RedirectChangePin = chkForceChangePIN.Checked;
				changed = true;
			}

            // Excluded tab
            string[] newExcluded = txtExcluded.Lines;
            Array.Sort(newExcluded);
            string[] oldExcluded = settings.ExcludedPaths;
            bool excChanged = false;
            if (newExcluded.Length != oldExcluded.Length) {
                excChanged = true;
            } else {
                Array.Sort(oldExcluded);
                if (newExcluded.Where((t, idx) => String.CompareOrdinal(t, oldExcluded[idx]) != 0).Any()) {
                    excChanged = true;
                }
            }
            if (excChanged) {
                settings.ExcludedPaths = newExcluded;
                changed = true;
            }

            if (drpIncludeIPs.SelectedIndex == 0 && settings.IncludeIPs) {
                settings.IncludeIPs = false;
                changed = true;
            } else if (drpIncludeIPs.SelectedIndex == 1 && !settings.IncludeIPs) {
                settings.IncludeIPs = true;
                changed = true;
            }
            newExcluded = txtExcludedIPs.Lines;
            List<IPAddressRange> lNewExcIPs = new List<IPAddressRange>(newExcluded.Length);
            lNewExcIPs.AddRange(newExcluded.Select(s => new IPAddressRange(s)).Where(r => r.IsValid));
            IPAddressRange[] newExcIPs = lNewExcIPs.ToArray();
            IPAddressRange[] oldExcIPs = settings.ExcludedIPs;
            excChanged = false;
            if (oldExcIPs == null) {
                if (newExcIPs.Length > 0) {
                    excChanged = true;
                }
            } else if (newExcIPs.Length != oldExcIPs.Length) {
                excChanged = true;
            } else {
                Array.Sort(newExcIPs);
                Array.Sort(oldExcIPs);
                if (newExcIPs.Where((t, idx) => t.CompareTo(oldExcIPs[idx]) != 0).Any()) {
                    excChanged = true;
                }
            }
            if (excChanged) {
                settings.ExcludedIPs = newExcIPs;
                changed = true;
            }

            return changed;
        }

        private string configFile;

        private static void ResetIIS() {
            Process proc = Process.Start("iisreset", "/noforce");
            if (proc != null) {
                proc.WaitForExit();
                MessageBox.Show("IIS has been restarted", "Restart IIS", MessageBoxButtons.OK);
            } else {
                MessageBox.Show("Unable to restart IIS. Please do so manually", "IISReset error", MessageBoxButtons.OK,
                                MessageBoxIcon.Error);
            }
        }

        public List<ServerSettings> Servers {
            get { return settings.Servers; }
        }

        public int ActiveServerId {
            get { return activeServerId; }
            set {
                if (value < 0 || value >= settings.ServerCount) return;
                activeServerId = value;
                UpdateServer();
            }
        }

        private void UpdateServer() {
            if (activeServerId < 0 || activeServerId >= settings.ServerCount) {
                activeServerId = 0;
            }
            ServerSettings server = settings.GetServer(activeServerId);
            txtServer.Text = server.Server;
            txtPort.Text = server.Port.ToString(CultureInfo.InvariantCulture);
            txtContext.Text = server.Context;
			chkSecure.Checked = server.Ssl;
			txtSecret.Text = server.Secret;
            chkSelfSigned.Checked = server.AcceptSelfSigned;
			txtProxyServer.Text = server.ProxyServer;
			txtProxyPort.Text = server.ProxyPort.ToString(CultureInfo.InvariantCulture);
			txtProxyContext.Text = server.ProxyContext;
			chkProxySecure.Checked = server.ProxySsl;
		}

        private void BtnServersClick(object sender, EventArgs e) {
            if (serverForm == null) {
                serverForm = new ServersForm(this) {Left = Right + 20, Top = Top + 10};
            }
            dirty = CheckServer();
            serverForm.SetList(settings.Servers, activeServerId);
            if (!serverForm.Visible) {
                serverForm.Show(this);
                btnServers.Text = "Update";
            }
            serverForm.Activate();
        }

        public void AddServer() {
            settings.Servers.Add(new ServerSettings());
            ActiveServerId = settings.ServerCount - 1;
        }

        public void DeleteServer(int selectedIndex) {
            if (selectedIndex <= 0 || selectedIndex >= settings.ServerCount) {
                return;
            }
            settings.Servers.RemoveAt(selectedIndex);
            serverForm.SetList(settings.Servers, 0);
        }

        private void txtOWAServer_Leave(object sender, EventArgs e) {
            if (!String.IsNullOrEmpty(txtOWAServer.Text)) {
                txtAuthUrl.Text = txtOWAServer.Text + PINsafeSettings.DEFAULT_AUTH_URL;
            }
        }

		private void CheckShowTuring(object sender, EventArgs e) {
			if (chkShowTuring.Checked) {
				chkShowPinpad.Checked = false;
				chkShowPinpad.Enabled = false;
			} else {
				chkShowPinpad.Enabled = true;
			}
			chkAutoGenerate.Enabled = (chkShowTuring.Checked || chkShowPinpad.Checked);
		}

		private void CheckShowPinpad(object sender, EventArgs e) {
			if (chkShowPinpad.Checked) {
				chkShowTuring.Checked = false;
				chkShowTuring.Enabled = false;
			} else {
				chkShowTuring.Enabled = true;
			}
			chkAutoGenerate.Enabled = (chkShowTuring.Checked || chkShowPinpad.Checked);
		}

		private void TxtServerChanged(object sender, EventArgs e) {
			if (String.IsNullOrEmpty(txtProxyServer.Text)) {
				txtProxyServer.Text = txtServer.Text;
			}
		}
    }
}