using System;
using System.Linq;
using System.Web;
using com.swivelsecure.pinsafe;

namespace com.swivelsecure.owafilter {

    public class PINsafeOWAFilter : IHttpModule {
        private PINsafeClient client;
        private PINsafeSettings settings;

        public void Init(HttpApplication context) {
            InitialiseSettings();
            context.BeginRequest += BeginRequest;
        }

        private void BeginRequest(object sender, EventArgs eventArgs) {
            HandleAuthentication((HttpApplication) sender);
        }

        private void InitialiseSettings() {
            client = PINsafeClient.GetWebInstance();
            settings = client.Settings;
        }

        private void HandleAuthentication(HttpApplication app) {
            HttpContext context = app.Context;
            HttpRequest request = context.Request;
            HttpCookie pinsafeCookie;
            string logonUrl = settings.LogonUrl;
            // Check whether this request requires authentication.
            bool authenticationRequired = CheckAuthenticationRequired(request.RawUrl, request.UserHostAddress);
            if (authenticationRequired) {
                // Authentication is required, check if the PINsafe cookie exists and is valid.
                pinsafeCookie = request.Cookies["PINSAFE"];
                bool cookieValid = false;
                if (pinsafeCookie != null) {
                        cookieValid = client.CookieValid(pinsafeCookie.Value);
                        if (!cookieValid) {
                            Debug.WriteLine("Authentication cookie is not valid");
                    }
                }
                // If not, redirect to the logon page.
                if (!cookieValid) {
                    context.Response.Redirect(logonUrl);
                } else {
                    // If it is valid, allow request to continue.
                    // Update the PINsafe cookie expiry time if necessary
                    string newCookieValue = client.UpdateCookie(pinsafeCookie.Value);
                    if (newCookieValue != null) {
                        pinsafeCookie = new HttpCookie("PINSAFE", newCookieValue) {Path = "/"};
                        context.Response.Cookies.Add(pinsafeCookie);
                    }
                }
            } else {
                Debug.WriteLine("Authentication not required for " + request.RawUrl);
            }
            string logoffUrl = settings.LogoffUrl.ToLower();
            if (!request.Url.OriginalString.ToLower().StartsWith(logoffUrl)) return;
            // Log off.
            pinsafeCookie = request.Cookies["PINSAFE"];
            if (pinsafeCookie == null) {
                return;
            }
            pinsafeCookie.Expires = new DateTime(1970, 1, 1);
            context.Response.Cookies.Remove("PINSAFE");
            context.Response.Cookies.Add(pinsafeCookie);
        }

        public void Dispose() {
            // Nothing to dispose.
        }

        private bool CheckAuthenticationRequired(string requestUrl, string clientIP) {
            if (!settings.FilterEnabled) {
                return false;
            }
            requestUrl = requestUrl.ToLower();
            return !settings.ExcludedPaths.Any(path => requestUrl.StartsWith(path.ToLower())) && new PINsafeClient(settings).CheckClientIP(clientIP);
        }
    }
}
